#include <EEPROM.h>
#include "RTClib.h"
#include <WiFi.h>
#include <WiFiClient.h>
#include <ezTime.h>
#include <TOTP.h>
#include <ErriezTTP229.h>
#include "Base32.h"
#include <FastLED.h>
#include <U8x8lib.h>

//******************************************************
#define ACCESS_POINT_NAME "<Your SSID>"        
#define ACCESS_POINT_PASSWORD "<Your Wifi Password>" 
#define HOSTNAME "myDoor"
#define masterPassword "012345AB72"
#define masterPwLength 10  // length of above code
#define localTimeZone "Australia/Adelaide"  // Best place on earth...

//External beeper is high to beep, Weigand readers usually low to beep
//Invert as necessary
#define beepOn 1
#define beepOff 0

//******************************************************

//Wiegand RFID Reader pins
#define dat0Pin 32
#define dat1Pin 33
#define beepPin 25
#define ledPin  26

#define wifiStatusLed 1
#define rtcStatusLed 0
#define lockPin 27
#define progPin 13

//Capacative touch keypad
#define TTP229_SDO_PIN     5
#define TTP229_SCL_PIN     4

//Fastled
#define NUM_LEDS 2
#define DATA_PIN 15
CRGB leds[NUM_LEDS];

Base32 base32;
RTC_DS3231 rtc;
Timezone myTZ;
WiFiClient espClient;

// Create keypad object
ErriezTTP229 ttp229;

//OLED Display
U8X8_SSD1306_128X32_UNIVISION_HW_I2C u8x8(/* reset=*/ U8X8_PIN_NONE);


volatile unsigned long tagID = 32;  //32 = bitwise 100000 - Set an initial mask to keep all tags 4 bytes long. WG26 (bits) plus another 6 = 32 bits or 4 bytes.
volatile unsigned long tmptagID;
volatile unsigned long lastBitArrivalTime; //Tracks Wiegand data timeout
volatile int bitCount = 0;

int buttonState = 0;
int rec_chips = 0;
int MAX_CARDS = 5;
byte CardBytes[4];
int MatchedTag = -1;

unsigned long lastClickMillis = 0;  //Tracks keypad pushes
volatile unsigned long lastFailure = 0; //Tracks wifi connection
boolean masterCodeActive = 0;
boolean screenCleared = 0;
boolean rtcInUse = 1;
char code[7] = {'0','0','0','0','0','0','\0'};  //Holds the 6 digit MFA code entered

char dateDayNow[3];
char dateDay[3];
char dateMonth[3];
char dateYear[3];

char timeHour[3];
char timeMinute[3];
char timeSecond[3];

char dateChr[9] = {'0','0','/','0','0','/','0','0','\0'};
char timeChr[9] = {'0','0',':','0','0',':','0','0','\0'};

char masterUnlockCode[masterPwLength +1];
int codeIndex = 0;

void ICACHE_RAM_ATTR ISRone(void)
{
  lastBitArrivalTime = millis();
  
  bitCount++;
  tagID <<= 1;
  tagID |= 1;
}
 
 
void ICACHE_RAM_ATTR ISRzero(void)
{
  lastBitArrivalTime = millis();
  bitCount++;
  tagID <<= 1;
}

void ICACHE_RAM_ATTR keyChange()
{
    // A key press changed
    ttp229.keyChange = true;
}

void syncTimeToRTC(){
  //Force the ezTime library to try internet time
  updateNTP();
  // time_status should now be either timeNeedsSync or timeSet 
  if (timeStatus() == timeSet){
    leds[rtcStatusLed] = CRGB::Green;  //Able to sync RTC with internet time within 10 sec.
	  leds[wifiStatusLed] = CRGB::Green;  //Internet access must be ok
    FastLED.show();
    rtc.adjust(DateTime(UTC.now())); //Set the time on the RTC
	  rtcInUse = 0;
  }else{
	  // Relying on RTC time
    leds[rtcStatusLed] = CRGB::DarkOrange;  
	  leds[wifiStatusLed] = CRGB::DarkOrange; //Connected to wifi but no internet or not able to sync time
    FastLED.show();
    DateTime now = rtc.now();
    UTC.setTime(now.unixtime());  //Set EZTime to whatever the RTC is.
	  rtcInUse = 1;
	  lastFailure = millis();
  }
}

void ConfigureWifi(){
  int timeout = 0;
  Serial.println("Configuring Wifi");
  
  WiFi.mode(WIFI_STA);
  WiFi.hostname(HOSTNAME);
  
  WiFi.begin (ACCESS_POINT_NAME, ACCESS_POINT_PASSWORD);

  while (WiFi.status() != WL_CONNECTED) {
    leds[wifiStatusLed] = CRGB::DarkOrange;
    FastLED.show();
    delay(250);
    leds[wifiStatusLed] = CRGB::Black;
    FastLED.show();
    delay(250);
    Serial.print(".");
    if (timeout == 20){  //10 Seconds
      Serial.println("");
      Serial.println("Unable to Connect to wifi");
      Serial.println("Will try to connect later");
      leds[wifiStatusLed] = CRGB::Red;  // No wifi
      leds[rtcStatusLed] = CRGB::DarkOrange;  // Internal RTC only
      FastLED.show();
      lastFailure = millis();
      syncTimeToRTC(); // It will fail but it will update the status led
      break;
    }
    timeout++;
  }
  if (WiFi.status() == WL_CONNECTED) {
	leds[wifiStatusLed] = CRGB::DarkOrange;
    //leds[wifiStatusLed] = CRGB::Green;
    FastLED.show();
    Serial.println("");
    Serial.println("WiFi connected");
    Serial.print("IP address: ");
    Serial.println(WiFi.localIP());
    syncTimeToRTC();
  }
}

void updateDay(){
  myTZ.dateTime("d").toCharArray(dateDay,3);
  myTZ.dateTime("m").toCharArray(dateMonth,3);
  myTZ.dateTime("y").toCharArray(dateYear,3);

  dateChr[0] = dateDay[0];
  dateChr[1] = dateDay[1];
  dateChr[3] = dateMonth[0];
  dateChr[4] = dateMonth[1];
  dateChr[6] = dateYear[0];
  dateChr[7] = dateYear[1];
}

void updateTime(){
  myTZ.dateTime("H").toCharArray(timeHour,3);
  myTZ.dateTime("i").toCharArray(timeMinute,3);
  myTZ.dateTime("s").toCharArray(timeSecond,3);

  timeChr[0] = timeHour[0];
  timeChr[1] = timeHour[1];
  timeChr[3] = timeMinute[0];
  timeChr[4] = timeMinute[1];
  timeChr[6] = timeSecond[0];
  timeChr[7] = timeSecond[1];
}

void setup() {
   // UC connections:
  pinMode(dat0Pin, INPUT);   // Data0
  pinMode(dat1Pin, INPUT);   // Data1
  pinMode(beepPin, OUTPUT);  // Beep
  pinMode(ledPin,  OUTPUT);  // Green LED
  pinMode(progPin, INPUT_PULLUP);   // Programming switch
  pinMode(lockPin, OUTPUT);   // To Transistor for lock output

  digitalWrite(beepPin, beepOff);// Turns off beep
  digitalWrite(ledPin,HIGH);// Turns off green led
  digitalWrite(lockPin,LOW); // Make sure the door is locked by default

  masterUnlockCode[masterPwLength] = '\0';
  
  Serial.begin(57600); // Serial Start
  EEPROM.begin(512);


  u8x8.begin(); //Start the OLED Display

  FastLED.addLeds<WS2812, DATA_PIN, GRB>(leds, NUM_LEDS);
  FastLED.setBrightness(32);

  //Initialise the RTC
  if (! rtc.begin()) {
    Serial.println("Couldn't find RTC!");
    leds[rtcStatusLed] = CRGB::Red;
    FastLED.show();
    while (1) delay(10);
  }else{
    leds[rtcStatusLed] = CRGB::DarkOrange;
    FastLED.show();
  }

  ConfigureWifi();

  // Set the Time Zone - Save TZData to EEPROM
  // TZdata is dependent on timezoned.rop.nl provided by the EZTime library
  // author. This may timebomb at some stage but it only effects the display
  // of the current time, not the time itself (that is pulled from pool.ntp.org)

  // 400 is EEPROM location, if you want to add more than 16 cards, you need to make this a higer number
  // Each card occupies 24 bytes of EEPROM, hence starting at the round number of 400
  
  if(!myTZ.setCache(400)) myTZ.setLocation(localTimeZone);

  // --- Data0 / Green Wire ----------------------------
  digitalWrite(dat0Pin, HIGH);  // Enable pull-up resistor
  attachInterrupt(dat0Pin, ISRzero, FALLING);
  //----------------------------------------------------
 
  // --- Data1 / White Wire ----------------------------
  digitalWrite(dat1Pin, HIGH);  // Enable pull-up resistor
  attachInterrupt(dat1Pin, ISRone,  FALLING);
  //----------------------------------------------------

  // Initialize keypad
  ttp229.begin(TTP229_SCL_PIN, TTP229_SDO_PIN);

  // Initialize interrupt pin on SD0
  attachInterrupt(digitalPinToInterrupt(TTP229_SDO_PIN), keyChange, FALLING);

  tagID = 32;
  bitCount = 0;

  updateTime();
  updateDay();
  
  // Screen Test...
  u8x8.setFont(u8x8_font_8x13B_1x2_f);
  u8x8.drawString(4,0,dateChr);
  u8x8.setFont(u8x8_font_px437wyse700a_2x2_n);
  u8x8.drawString(0,2,timeChr);
}  

void loop() {
  events(); // Process EZtime

  //Update the display
  if (minuteChanged()){
    updateTime();
    u8x8.setFont(u8x8_font_px437wyse700a_2x2_n);
    u8x8.drawString(0,2,timeChr);

    myTZ.dateTime("d").toCharArray(dateDayNow,2);
    if (dateDay != dateDayNow){
      strcpy(dateDay, dateDayNow);
      u8x8.setFont(u8x8_font_8x13B_1x2_f);
      u8x8.drawString(4,0,dateChr);
    }
  }

  if (secondChanged()){
    updateTime();
    u8x8.setFont(u8x8_font_px437wyse700a_2x2_n);
    u8x8.drawString(0,2,timeChr);
  }

  if (screenCleared == 1){
    u8x8.setFont(u8x8_font_8x13B_1x2_f);
    u8x8.drawString(4,0,dateChr);
    screenCleared = 0;
  }

  
  if(bitCount > 0 && millis() - lastBitArrivalTime >  250){
    //convert the 32 bits into 4 byte arrays
    CardBytes[0] = lowByte(tagID);
    CardBytes[1] = highByte(tagID);
    tmptagID = tagID >> 16;
    CardBytes[2] = lowByte(tmptagID);
    CardBytes[3] = highByte(tmptagID);
    
    buttonState = digitalRead(progPin);
    
    if(buttonState == 0){
      //Programming mode
      WriteChip();
    }else {
      //Normal mode
      rec_chips = 0;
      Serial.print(CardBytes[0],HEX);
      Serial.print(CardBytes[1],HEX);
      Serial.print(CardBytes[2],HEX);
      Serial.println(CardBytes[3],HEX);
      MatchedTag = Checkchip(CardBytes);
      if (MatchedTag != -1){
        Serial.print("Matched Tag: ");
        Serial.println(MatchedTag);

        if (masterCodeActive == 0){ //Using OTP unless master code button was pushed first
          if (checkOTP(MatchedTag) == 1){
            unlockDoor();  //Tag and OTP match
            MatchedTag = -1;
          }else{
            denyAccess(); //Tag matched but OTP didn't
            MatchedTag = -1;
          }
        }
      }else{
        denyAccess(); // Tag did not match any recorded
      }
    }
  tagID = 32;
  bitCount = 0;
  }

  if(ttp229.keyChange == true) {
    lastClickMillis = millis(); //Reset the timer
    uint8_t tempInt = ttp229.GetKey16();
    if(tempInt != 0){ // Only button press events, not button release
      digitalWrite(beepPin,beepOn);// Turns on beep - Helps with debouncing
      delay (100);
      digitalWrite(beepPin,beepOff);// Turns off beep
      
      if (tempInt == 13) { // "*"
        masterCodeActive = 1; 
      }else{
        if (MatchedTag != -1 && masterCodeActive == 1){
          masterUnlockCode[codeIndex] = convertToChar(tempInt); //Add pressed key to array storing master code attempt
          //Serial.print(codeIndex);
          //Serial.print(":");
          //Serial.print(convertToChar(tempInt));
        
          if (codeIndex == masterPwLength -1){ //Array's are zero indexed
            Serial.println("");
            Serial.println("Master Pw check");
            if (strcmp(masterUnlockCode, masterPassword) == 0) {
              Serial.println("Matched");
              unlockDoor();
            }else{
              Serial.print(masterUnlockCode);
              Serial.print(" Didn't match ");
              Serial.println(masterPassword);
            }
            codeIndex=0;
            MatchedTag = -1;
            masterCodeActive = 0;
          }
          codeIndex++;
        }
      }
    }
    ttp229.keyChange = false;
  }
  if (millis() - lastClickMillis > 5000){ //More than 5 seconds have passed, clear out codes
    codeIndex=0;
    MatchedTag = -1;
    masterCodeActive = 0;
  }

  // Wifi connected but time is not in sync
  if (WiFi.status() == WL_CONNECTED && millis() - lastFailure > 30000 && rtcInUse == 1) {  //Five Minutes
    syncTimeToRTC();
  }
  if (WiFi.status() == WL_CONNECTED && rtcInUse == 0 && timeStatus() == timeNeedsSync) syncTimeToRTC(); //This ensures the LEDs are updated if something changes.

    // Wifi not connected - Try to reconnect and sync time.
  if (WiFi.status() != WL_CONNECTED && millis() - lastFailure > 60000) {  //10 Mins
    ConfigureWifi(); //Also calls syncTimeToRTC
  }
}

void WriteChip(){
  // initialise internal pseudo-random number generator
  randomSeed(analogRead(14));
    
  char *validChars = "1234567890ABCDEFGHIJKLMNOPQRSTUVWXYZ";
  char mfaSecret[20];
  byte mfaSecretB[20];
  byte* tempEncoded = NULL;
  char encodedSecret[33];
  encodedSecret[32] = '\0';
    
  if (rec_chips < MAX_CARDS){
    EEPROM.begin(512);
    Serial.print("Added Tag: ");
    for(int i = 0; i < 4; i++){
      EEPROM.write(4*rec_chips + i, CardBytes[i]);
      Serial.print(CardBytes[i],HEX);
    }
    Serial.println("");

    //Generate MFA Secret for the tag..
    for (int n=0;n<20;n++){
      mfaSecret[n] = validChars[random(0, 36)]; //Pick a random character from validChars
      mfaSecretB[n] = mfaSecret[n];
      //Serial.print(char(mfaSecret[n]));
      EEPROM.write(4*MAX_CARDS+n+(rec_chips*20),mfaSecret[n]);
    }
    Serial.println("");

    //Base32 Encode the secret
    base32.toBase32(mfaSecretB, 20, (byte*&)tempEncoded, true);

    //Load it into a null terminated array
    for(int i=0;i<32;i++){
      encodedSecret[i] = char(tempEncoded[i]);
    }
    
    Serial.print("Generated Secret for tag ");
    Serial.print(rec_chips);
    Serial.print(" ");
    Serial.println(encodedSecret);
	  //String urlString1 = "http://api.qrserver.com/v1/create-qr-code/?color=000000&bgcolor=FFFFFF&data=otpauth%3A%2F%2Ftotp%2FRFID%2520MFA%2520Access%3ATag%2520";
	  //String urlString2 = "%3Falgorithm%3DSHA1%26digits%3D6%26period%3D30%26secret%3D";
	  //String urlString3 = "&qzone=1&margin=0&size=400x400&ecc=L";
	  //String tempStr1 = String(urlString1 + rec_chips + urlString2 + encodedSecret + urlString3);
	  String urlString1 = "http://api.qrserver.com/v1/create-qr-code/?data=otpauth%3A%2F%2Ftotp%2FRFID%2520MFA%2520Access%3ATag%2520";
	  String urlString2 = "%3Falgorithm%3DSHA1%26digits%3D6%26period%3D30%26secret%3D";
	  String tempStr1 = String(urlString1 + rec_chips + urlString2 + encodedSecret);
	  Serial.print("QR Code Link: ");
	  Serial.println(tempStr1);

    //Write the secret to EEPROM
    EEPROM.commit();
  }
  Serial.println("");
  ++rec_chips;
}
  
int Checkchip(byte TagData[]){
  EEPROM.begin(512);
  int count = 0;
  byte tmpByte[4];
  while (count < MAX_CARDS){
    for(int i = 0; i < 4; i++){
      tmpByte[i] = EEPROM.read(4*count + i);
       //Serial.print(tmpByte[i], HEX);
       //Serial.print(" - ");
       //Serial.println(TagData[i], HEX);
    }
      if (tmpByte[0] == TagData[0] && tmpByte[1] == TagData[1] && tmpByte[2] == TagData[2] && tmpByte[3] == TagData[3])
        {
        // We matched all 4 bytes of data to one in EEPROM
        return count;
        }
      else
        {
          //Serial.print("Did not match tag : ");
          //Serial.println(count);
          count++;
        }
  }
 // If we got here, no tags matched - Access denied!
 //Serial.println("No match found"); 
 return -1; 
}

int checkOTP(int MatchedTag){
  EEPROM.begin(512);
  codeIndex = 0;
  lastClickMillis = millis(); //Reset the timer
  Serial.println("Checking OTP");
  u8x8.clear();
  u8x8.setFont(u8x8_font_courB24_3x4_f);
  u8x8.drawString(3,0,"OTP");
  while (millis() - lastClickMillis < 5000){
    if (ttp229.keyChange == true) {
      lastClickMillis = millis(); //Reset the timer

      uint8_t tmpKey = ttp229.GetKey16();
      if(tmpKey != 0){
        code[codeIndex] = convertToChar(tmpKey);
        digitalWrite(beepPin,beepOn);// Turns on beep - Helps debounce
        delay (100);
        digitalWrite(beepPin,beepOff);// Turns off beep
        Serial.print(convertToChar(tmpKey));
        codeIndex = codeIndex + 1;
      }
      
      if (codeIndex == 6) {
        Serial.println("");
        uint8_t mfaSecret[20];

        // Read the RAW secret from EEPROM associated with that tag
        for (int n=0;n<20;n++){
          mfaSecret[n] = EEPROM.read(4*MAX_CARDS+n+(MatchedTag*20));
          //Serial.print(char(mfaSecret[n]));
        }
        //Serial.println("");
        TOTP totp = TOTP(mfaSecret, 20); //Set the object to the current secret
        DateTime now = rtc.now(); //Get the time from RTC

         // Get the code for 30 seconds in the past, now, and 30 seconds in the future,
         // This allows us to have some time skew but still get access


        String totP = String(totp.getCode(now.unixtime()-30));
        String totN = String(totp.getCode(now.unixtime()));
        String totF = String(totp.getCode(now.unixtime()+30));

//        Serial.print(now.hour());
//        Serial.print(":");
//        Serial.print(now.minute());
//        Serial.print(":");
//        Serial.println(now.second());
        
        Serial.println(totP);
        Serial.println(totN);
        Serial.println(totF);
        

        if (strcmp(code, totN.c_str()) == 0 || strcmp(code, totP.c_str()) == 0 || strcmp(code, totF.c_str()) == 0) { //Compare the code entered with the generated one.
          return 1;  //Access Allowed
        }else{
          return 0; //Access Denied
        }
        memset(code, 0, 6); //Clear the array
      }
      ttp229.keyChange = false;
    }
  } 
  //Timed out 
  return 0;
}

void unlockDoor(){
  u8x8.clear();
  u8x8.setFont(u8x8_font_courB24_3x4_f);
  u8x8.drawString(4,0,"OK");
  digitalWrite(lockPin,HIGH); // Unlock the door
  digitalWrite(ledPin,LOW); // Turns on green led
  delay (200);
  digitalWrite(beepPin,beepOn);// Turns on beep
  delay (50);
  digitalWrite(beepPin,beepOff);// Turns off beep
  delay (50);
  digitalWrite(beepPin,beepOn);// Turns on beep
  delay (50);
  digitalWrite(beepPin,beepOff);// Turns off beep
  //Leave the door unlocked for a bit...     
  delay (5000);
  digitalWrite(ledPin,HIGH); // Turns off green led
  digitalWrite(lockPin,LOW); // Lock the door again
  u8x8.clear();
  screenCleared = 1;
}

void denyAccess(){
  Serial.println("Access denied");
  u8x8.clear();
  u8x8.setFont(u8x8_font_courB24_3x4_f);
  u8x8.drawString(2,0,"DENY");
  digitalWrite(beepPin,beepOn);// Turns on beep
  delay (1000);
  digitalWrite(beepPin,beepOff);// Turns off beep
  u8x8.clear();
  screenCleared = 1;
}



char convertToChar(int charNumber){
  switch (charNumber) {
    case 1:
      return '1';
      break;
    case 2:
      return '2';
      break;
    case 3:
      return '3';
      break;
    case 5:
      return '4';
      break;
    case 6:
      return '5';
      break;
    case 7:
      return '6';
      break;
    case 9:
      return '7';
      break;
    case 10:
      return '8';
      break;
    case 11:
      return '9';
      break;
    case 13:
      return '*';
      break;
    case 14:
      return '0';
      break;
    case 15:
      return '#';
      break;
    case 4:
      return 'A';
      break;
    case 8:
      return 'B';
      break;
    case 12:
      return 'C';
      break;
    case 16:
      return 'D';
      break;
    default:
      return '\0';
      break;
  }
}
